#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include "common.h"

// the offset indicates...
// 012
// 345
// 678
int offset_x[] = {-1, -1, -1,  0,  0,  0, +1, +1, +1};
int offset_y[] = {-1,  0, +1, -1,  0, +1, -1,  0, +1};

// stuff
int bs; // bin size
int np; // number of max particles per bin
double cutoff;

// arrays
#define b(x, y, p, n) (bins[(x)+(y)*(n)+(p)*(n)*(n)])
#define c(x, y, n) (counts[(x)+(y)*(n)])

int* bins;
int* counts;

void assign_bins(double cutoff, int n, particle_t* particles)
{
  for (int p=0; p<n; p++)
  {
    int x = (int)(particles[p].x/cutoff);
    int y = (int)(particles[p].y/cutoff);

    b(x, y, c(x, y, bs)++, bs) = p;

    if (c(x, y, bs) > np)
    {
      printf("error!\n");
      exit(-1);
    }
  }
}

//
//  benchmarking program
//
int main(int argc, char **argv)
{    
  if (find_option(argc, argv, "-h") >= 0)
  {
    printf("Options:\n");
    printf("-h to see this help\n");
    printf("-n <int> to set the number of particles\n");
    printf("-o <filename> to specify the output file name\n");
    return 0;
  }

  int n = read_int(argc, argv, "-n", 1000);

  char *savename = read_string(argc, argv, "-o", NULL);

  FILE *fsave = savename ? fopen(savename, "w") : NULL;
  particle_t *particles = (particle_t*) malloc(n * sizeof(particle_t));
  set_size(n);
  init_particles(n, particles);

  bs = get_binsize();
  np = bs / 10;
  cutoff = get_cutoff();

  bins = (int*)malloc(bs*bs*np*sizeof(int));
  counts = (int*)malloc(bs*bs*sizeof(int));
  memset(counts, 0, bs*bs*sizeof(int));

  //
  //  simulate a number of time steps
  //
  double simulation_time = read_timer();
  for (int step = 0; step < NSTEPS; step++)
  {
    assign_bins(cutoff, n, particles);

    //
    //  clear forces
    //
    for (int i = 0; i < n; i++)
    {
      particles[i].ax = particles[i].ay = 0;
    }

    for (int bx=0; bx<bs; bx++)
    {
      for (int by=0; by<bs; by++)
      {
        if (c(bx, by, bs) != 0)
        {
          for (int o=0; o<9; o++)
          {
            int nx = bx+offset_x[o];
            int ny = by+offset_y[o];

            if (nx < 0 || nx >= bs) continue;
            if (ny < 0 || ny >= bs) continue;

            if (c(nx, ny, bs) != 0)
            {
              for (int i=0; i<c(bx, by, bs); i++)
              {
                for (int j=0; j<c(nx, ny, bs); j++)
                {
                  apply_force(particles[b(bx, by, i, bs)], particles[b(nx, ny, j, bs)]);
                }
              }
            }
          }
        }
      }
    }

    //
    //  move particles
    //
    for (int i = 0; i < n; i++) 
      move(particles[i]);

    //
    //  save if necessary
    //
    if (fsave && (step%SAVEFREQ) == 0)
      save(fsave, n, particles);

    memset(counts, 0, bs*bs*sizeof(int));
  }
  simulation_time = read_timer() - simulation_time;

  printf("n = %d, simulation time = %g seconds\n", n, simulation_time);

  free(particles);
  if (fsave)
    fclose(fsave);

  return 0;
}
